
local tooltippy

do
  local tooltip_rim = CreateFrame("Frame")
  tooltip_rim:SetLayer(500)

  local tooltip_bg = CreateFrame("Frame", tooltip_rim)

  local tooltip = CreateFrame("Text_Multiline", tooltip_bg)
  tooltip:SetWidth(280)
  tooltip:SetSize(0.65)
  tooltip:SetText("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum")
  tooltip:SetColor(1, 1, 1)
  tooltip:SetPoint("CENTER", UIParent, "CENTER")
  tooltip:ForceHeight() -- urgh

  tooltip_rim:Hide()

  local last_mouse_x = nil
  local last_mouse_y = nil
  local hide_on_mmove = false
  
  function tooltip:Tick()
    local mx, my = GetMouseX(), GetMouseY()
    
    if hide_on_mmove and (last_mouse_x ~= mx or last_mouse_y ~= my) then
      tooltip_rim:Hide()
    end

    tooltip:ClearAllPoints(true, false)
    tooltip:SetPoint("BOTTOMRIGHT", UIParent, "TOPLEFT", mx - 5, my - 5)
    
    local clamp_up = false
    local clamp_right = false
    if tooltip:GetTop() < 5 then clamp_up = true end
    if tooltip:GetLeft() < 5 then clamp_right = true end
    
    tooltip:ClearAllPoints(true, false)
    if clamp_up then
      tooltip:SetPoint("TOP", UIParent, "TOP", nil, 5)
    else
      tooltip:SetPoint("BOTTOM", UIParent, "TOP", nil, my - 5)
    end
    if clamp_right then
      tooltip:SetPoint("LEFT", UIParent, "LEFT", 5, nil)
    else
      tooltip:SetPoint("RIGHT", UIParent, "LEFT", mx - 5, nil)
    end
  end

  tooltip_bg:SetPoint("TOPLEFT", tooltip, "TOPLEFT", -3, -3)
  tooltip_bg:SetPoint("BOTTOMRIGHT", tooltip, "BOTTOMRIGHT", 3, 3)
  tooltip_bg:SetBackgroundColor(0, 0, 0.3)

  tooltip_rim:SetPoint("TOPLEFT", tooltip_bg, "TOPLEFT", -2, -2)
  tooltip_rim:SetPoint("BOTTOMRIGHT", tooltip_bg, "BOTTOMRIGHT", 2, 2)
  tooltip_rim:SetBackgroundColor(0.4, 0.4, 1.0)

  function tooltippy(frame, func)
    local omin, omout = frame.MouseIn, frame.MouseOut
    function frame:MouseIn()
      if omin then omin(self) end
      
      local tex = func()
      if tex then
        tooltip_rim:Show()
        tooltip:SetText(tex)
        tooltip:ForceHeight() -- urgh
        hide_on_mmove = false
      end
    end
    function frame:MouseOut()
      if omout then omout(self) end
      
      hide_on_mmove = true
      last_mouse_x, last_mouse_y = GetMouse()
    end
  end
end



loadfile("terra.lua")()
loadfile("worldparams.lua")()

local worldx = 44
local worldy = 40

local terra = terra_init(mappieces, entities)
terra.load(assert(io.snatch("level_worldmap.lev.lua"), lev))
local gameworld = CreateFrame("Frame")
gameworld:SetCoordinateScale(0, 0, (worldx))

gameworld:SetWidth(worldx * 16)
gameworld:SetHeight(worldy * 16)

gameworld:SetPoint("CENTER", UIParent, "CENTER", 120, 0)
gameworld:SetBackgroundColor(0.5, 0.5, 0.5)

function gameworld:Draw()
  terra.render()
end




--[[
local wordz = {
  "hello",
  "goodbye",
  "waffles",
  "I am a ludafisk",
  "meow meow meow",
  "beforth, I am was",
    "hello",
  "goodbye",
  "waffles",
  "I am a ludafisk",
  "meow meow meow",
  "beforth, I am was",
    "hello",
  "goodbye",
  "waffles",
  "I am a ludafisk",
  "meow meow meow",
  "beforth, I am was",
    "hello",
  "goodbye",
  "waffles",
  "I am a ludafisk",
  "meow meow meow",
  "beforth, I am was",
}

local scrolly = CreateFrame("Scroll")
scrolly:SetPoint("CENTER", UIParent, "CENTER", 10, 10)
scrolly:SetWidth(300)
scrolly:SetHeight(90)
scrolly:SetBackgroundColor(0.1, 0.1, 0.1)

local last = CreateFrame("Frame")
last:SetPoint("BOTTOMLEFT", scrolly.internal, "TOPLEFT")

for _, v in ipairs(wordz) do
  local wbut = CreateFrame("Button", scrolly)
  
  local word = CreateFrame("Text", wbut)
  word:SetText(v)
  word:SetColor(1, 1, 1)
  word:SetPoint("TOPLEFT", last, "BOTTOMLEFT")
  
  function wbut:Click()
    print("Clicked", v)
  end
  
  last = word
end

scrolly.internal:SetPoint("BOTTOM", last, "BOTTOM")
]]

local MakeEntity
do
  local dx = {0, 0, 1, -1}
  local dy = {1, -1, 0, 0}
  
  local function costize(token)
    if not token then return nil end
    if token:find("building") then return 10 end
    if token:find("road") then return 1 end
    if token:find("water") then return nil end  -- impassable
    assert(false)
  end
  
  function MakeEntity()
    local rv = CreateFrame("Frame", gameworld)
    impose_standard_ai(rv)
    function rv:Tick()
      self:TickStandard()
      
      self:SetPoint("CENTER", UIParent, "TOPLEFT", self.x + 0.5, self.y + 0.5)
    end
    
    rv:SetSize(0.5)
    rv:SetBackgroundColor(1, 0, 0)
    
    rv.x = 0
    rv.y = 0
    
    rv:Tick_Add(function (self)
      if self.Movement then self:Movement() end
    end)
    
    function rv:MoveTo(x, y, slowdown)
      print("Entering moveto")
      
      local sx, sy = math.round(self.x), math.round(self.y)
      
      local heep = {{x = sx, y = sy, c = 0}}
      
      local dest = {}
      
      local succeed = (function ()
        for pos in function() return heap_extract(heep) end do
          if pos.x == x and pos.y == y then return true end -- found it!
          
          for i = 1, #dx do
            local nx, ny = pos.x + dx[i], pos.y + dy[i]
            
            local sp, ep = costize(terra.layers[1][pos.x][pos.y]), costize(terra.layers[1][nx][ny])
            if sp and ep then
              local cost = math.max(sp, ep)
              
              if not (dest[nx] and dest[nx][ny] and dest[nx][ny].c <= pos.c + cost) then
                if not dest[nx] then dest[nx] = {} end
                dest[nx][ny] = {fromx = pos.x, fromy = pos.y, c = pos.c + cost, slowth = cost}
                heap_insert(heep, {x = nx, y = ny, c = pos.c + cost})
              end
            end
          end
        end
        
        return false
      end)()
      
      if succeed then
        local path = {}
        local cx, cy = x, y
        print(cx, cy, x, y, self.x, self.y)
        while cx ~= sx or cy ~= sy do
          table.insert(path, {x = cx, y = cy, slowth = dest[cx][cy].slowth})
          cx, cy = dest[cx][cy].fromx, dest[cx][cy].fromy
        end
        
        local pathreverse = {}
        for n = #path, 1, -1 do
          table.insert(pathreverse, path[n])
        end
        
        self.Movement = coroutine.wrap(function ()
          for _, v in ipairs(pathreverse) do
            local cx, cy = self.x, self.y
            local dx, dy = v.x, v.y
            local delay = v.slowth * 10 * (slowdown or 1)
            for tim = 1, delay do
              self.x, self.y = lerp(tim / delay, cx, dx), lerp(tim / delay, cy, dy)
              coroutine.yield()
            end
          end
          
          self.Movement = nil
        end)
      else
        print("Couldn't move to", x, y)
      end
      
      print("Exiting moveto")
    end
    
    rv.status = "Waiting."
    
    function rv:SetStatus(stat)
      self.status = stat
    end
    
    tooltippy(rv, function ()
      return rv.status
    end)
    
    return rv
  end
end

local tasks = {}
-- subtask: {x, y, status, wait, status_wait, task}

local MakeRunner
do
  function MakeRunner()
    local rn = MakeEntity()
    
    rn:Tick_AddCoro(function ()
      while true do
        coroutine.yield()
        
        if #tasks > 0 then
          local task = tasks[1]
          table.remove(tasks, 1)
          
          for _, subtask in ipairs(task) do
            rn:SetStatus(subtask.status)
            rn:MoveTo(subtask.x, subtask.y)
            while rn.Movement do
              coroutine.yield()
            end
            
            rn:SetStatus(subtask.status_wait)
            coroutine.pause(subtask.wait)
            
            subtask.task()
          end
        else
          if not rn.Movement then
            if math.random(2) == 1 then
              rn:SetStatus("Getting distracted.")
              coroutine.pause(120)
            else
              rn:SetStatus("Wandering aimlessly.")
              rn:MoveTo(math.random(worldx) - worldx / 2, math.random(worldy) - worldy / 2, 2)
            end
          end
        end
      end
    end)
  end
end

local select_width = 1
local select_height = 1
local cursor_mode = nil
local warehouse = nil
local depots = {}
local function ResetButton()
  select_width = 1
  select_height = 1
  cursor_mode = nil
end
local function StandardSelectButton(frame, ident, func)
  function frame:Click()
    if cursor_mode == ident or func() then
      ResetButton()
    else
      cursor_mode = ident
    end
  end
end

local warehouse_button = CreateFrame("Button")
local warehouse_text = CreateFrame("Text", warehouse_button)
warehouse_text:SetColor(1, 1, 1)
warehouse_text:SetText("Build Warehouse")
warehouse_text:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 40, 40)
StandardSelectButton(warehouse_button, "warehouse", function ()
  if warehouse then
    return true
  end
  select_width = 2
  select_height = 2
end)

local runner_button = CreateFrame("Button")
local runner_text = CreateFrame("Text", runner_button)
runner_text:SetColor(1, 1, 1)
runner_text:SetText("Hire Runner")
runner_text:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 40, 80)
runner_button.Click = MakeRunner

local function typetest(x, y, typ)
  local str = terra.layers[1][x][y]
  if not str then return end
  return str:find(typ)
end

local depot_button = CreateFrame("Button")
local depot_text = CreateFrame("Text", depot_button)
depot_text:SetColor(1, 1, 1)
depot_text:SetText("Build Depot")
depot_text:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 40, 120)
StandardSelectButton(depot_button, "depot", function ()
  select_width = 1
  select_height = 1
end)

local function typetest(x, y, typ)
  local str = terra.layers[1][x][y]
  if not str then return end
  return str:find(typ)
end

local docks = {}
for x = -worldx / 2, worldx / 2 do
  for y = -worldy / 2, worldy / 2 do
    if typetest(x, y, "building") and (typetest(x+1, y, "water") or typetest(x-1, y, "water") or typetest(x, y+1, "water") or typetest(x, y-1, "water")) then
      table.insert(docks, {x = x, y = y})
    end
  end
end

local function warehouse_tick()
  while warehouse.contained + warehouse.reserved_in + 1 <= warehouse.capacity do
    local tasklist = {}
    
    local dock = docks[math.random(#docks)]
    
    table.insert(tasklist, {x = dock.x, y = dock.y, status = "Going to buy some drugs", wait = math.random() * 240 + 240, status_wait = "Trying to buy drugs", task = function () end})
    table.insert(tasklist, {x = warehouse.x, y = warehouse.y, status = "Dropping cargo off", wait = 240, status_wait = "Dropping cargo off", task = function ()
      warehouse.contained = warehouse.contained + 1
      warehouse.reserved_in = warehouse.reserved_in - 1
    end})
    
    table.insert(tasks, tasklist)
    
    warehouse.reserved_in = warehouse.reserved_in + 1
  end
end
local function depot_tick(self)
  local amount_each = 0.2
  if self.contained + self.reserved_in + amount_each <= self.capacity and warehouse and warehouse.contained - warehouse.reserved_out >= amount_each then
    local tasklist = {}
    
    warehouse.reserved_out = warehouse.reserved_out + amount_each
    self.reserved_in = self.reserved_in + amount_each
    
    table.insert(tasklist, {x = warehouse.x, y = warehouse.y, status = "Picking cargo up", wait = 240, status_wait = "Picking cargo up", task = function ()
      warehouse.contained = warehouse.contained - amount_each
      warehouse.reserved_out = warehouse.reserved_out - amount_each
    end})
    table.insert(tasklist, {x = self.x, y = self.y, status = "Dropping cargo off", wait = 240, status_wait = "Dropping cargo off", task = function ()
      self.contained = self.contained + amount_each
      self.reserved_in = self.reserved_in - amount_each
    end})
    
    table.insert(tasks, tasklist)
  end
  
  self.contained = math.max(self.contained - 0.0002, 0)
end



local highlight = CreateFrame("Frame", gameworld)
highlight:SetBackgroundColor(1, 1, 1, 0.8)
highlight:SetWidth(1)
highlight:SetHeight(1)
highlight:SetLayer(5)



local taski = {}

function tick_loop(...)
  local x, y = GetMouse()
  
  local tx, ty = gameworld:GetCenter()
    
  local dx, dy = x - tx, y - ty
  dx, dy = math.floor(dx / 16 + 0.5 - select_width / 2), math.floor(dy / 16 + 0.5 - select_height / 2)
  
  highlight:SetPoint("TOPLEFT", UIParent, "TOPLEFT", dx, dy)
  highlight:SetSize(select_width, select_height)
end
function key(ki, _, event)
  if ki == "mouse_left" and event == "press" then
    print("cursor mode:", cursor_mode)
    
    if cursor_mode == "warehouse" then
      local x, y = highlight:GetTopLeft()
      
      if typetest(x, y, "building") and typetest(x + 1, y, "building") and typetest(x, y + 1, "building") and typetest(x + 1, y + 1, "building") then
        warehouse = CreateFrame("Texture", gameworld)
        warehouse:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
        warehouse:SetTexture("warehouse")
        warehouse:SetSize(2)
        tooltippy(warehouse,
          function () return ("This is your drug warehouse.\n\nDrugs on-hand: %f"):format(warehouse.contained) end
        )
        warehouse.Tick = warehouse_tick
        warehouse.contained = 0
        warehouse.reserved_out = 0
        warehouse.reserved_in = 0
        warehouse.capacity = 4 + #depots * 4
        warehouse.x = x + 1
        warehouse.y = y + 1
        ResetButton()
      end
    end
    
    if cursor_mode == "depot" then
      local x, y = highlight:GetTopLeft()
      
      if typetest(x, y, "building") then
        local depot = CreateFrame("Texture", gameworld)
        depot:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
        depot:SetTexture("warehouse")
        depot:SetSize(1)
        tooltippy(depot,
          function () return ("This is a drug depot.\n\nDrugs on-hand: %f"):format(depot.contained) end
        )
        depot.Tick = depot_tick
        depot.contained = 0
        depot.reserved_out = 0
        depot.reserved_in = 0
        depot.capacity = 1
        depot.x = x
        depot.y = y
        if warehouse then warehouse.capacity = warehouse.capacity + 4 end
        ResetButton()
        
        table.insert(depots, depot)
      end
    end
  end
  
  if ki == "mouse_right" and event == "press" then
    ResetButton()
  end
end
-- think there's more